
import json
import platform
import os
import shutil
import stat
import subprocess
import sys
import tarfile

import ssl
ssl._create_default_https_context = ssl._create_unverified_context

try:
    from urllib.request import urlretrieve, urlopen
except:
    # Python 2
    from urllib import urlretrieve
    from urllib2 import urlopen

from modian import PREFIXES

def conda_package_url(name, version=None, label="main"):
    system = platform.system().lower()
    machine = platform.machine()
    if system == "windows":
        system = "win"
        machine = "x86_64" if machine == "AMD64" else "x86"
    fd = urlopen("http://api.anaconda.org/package/{}".format(name))
    data = json.load(fd)
    fd.close()
    if version is None:
        version = data["latest_version"]
    b = None
    for f in data["files"]:
        if f["version"] != version:
            continue
        if label not in f["labels"]:
            continue
        if f["attrs"]["operatingsystem"] is not None:
            if f["attrs"]["operatingsystem"] != system and f["attrs"]["platform"] != system:
                continue
            if f["attrs"]["machine"] != machine and f["attrs"]["arch"] != machine :
                continue
        if b is None or f["attrs"]["build_number"] > b["attrs"]["build_number"]:
            b = f
    return "http:{}".format(b["download_url"]) if b else None

def prepare_dest(dest):
    destdir = os.path.dirname(dest)
    if not os.path.exists(destdir):
        os.makedirs(destdir)

def conda_package_extract(conda_url, prefix):
    print("downloading {}".format(conda_url))
    localfile = urlretrieve(conda_url)[0]
    fmt = conda_url.split(".")[-1]
    def match_member(m):
        return m.name.split('/')[0] != 'info'
    with tarfile.open(localfile, "r:%s"%fmt) as tar:
        for m in tar:
            if match_member(m):
                dest = os.path.join(prefix, m.name)
                print("installing %s" % dest)
                tar.extract(m, prefix)
    os.unlink(localfile)

def is_installed(progname):
    if sys.version_info[0] < 3:
        paths = os.environ["PATH"].split(":")
        for p in paths:
            if os.path.exists(os.path.join(p, progname)):
                return True
        return False
    print(shutil.which(progname))
    return shutil.which(progname) is not None

def installation_prefix():
    if platform.system() == "Windows":
        return PREFIXES[0]
    if os.getuid() == 0:
        return PREFIXES[0]
    else:
        return PREFIXES[-1]

def setup(*specs):
    in_conda = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("-f", "--force", default=False, action="store_true",
            help="Force installation")
    args = parser.parse_args()
    prefix = installation_prefix()
    for spec in specs:
        name = spec["pkg"].split('/')[-1]
        if not args.force:
            print("# checking for {}".format(name))
            skip = True
            for prog in spec.get("check_progs", []):
                if not is_installed(prog):
                    skip = False
                    break
            if skip and "check_install" in spec:
                skip = spec["check_install"]()
            if skip:
                print("# {} is already installed.".format(name))
                continue
        print("# installing {} in {}".format(spec["pkg"], prefix))
        if "install" in spec:
            spec["install"](prefix)
            continue
        if in_conda:
            subprocess.call(["conda", "install", spec["pkg"].split("/")[-1]])
            continue
        pkg = conda_package_url(spec["pkg"], version=spec.get("version"))
        if pkg is None:
            print("Error: no package found for your system!")
            continue
        conda_package_extract(pkg, prefix)

def nusmv_install(prefix):
    def chmod_x(filename):
        st = os.stat(filename)
        os.chmod(filename, st.st_mode | stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH)

    def prepare_dest(dest):
        destdir = os.path.dirname(dest)
        if not os.path.exists(destdir):
            os.makedirs(destdir)

    def install_bin_from_tar(tar, member, dest):
        fd = tar.extractfile(member)
        print("installing %s" % dest)
        prepare_dest(dest)
        with open(dest, "wb") as o:
            o.write(fd.read())
        fd.close()
        chmod_x(dest)

    def install_from_tarurl(url, match_member, prefix):
        localfile = urlretrieve(url)[0]
        if url.endswith(".gz"):
            mode = "r:gz"
        elif url.endswith(".bz2"):
            mode = "r:bz2"
        with tarfile.open(localfile, mode) as t:
            for m in t:
                if match_member(m):
                    dest = os.path.join(prefix, "bin", os.path.basename(m.name))
                    install_bin_from_tar(t, m, dest)
        os.unlink(localfile)

    system = platform.system().lower()
    if system == "linux":
        system = "%s%s" % (system, platform.architecture()[0][:2])
    #url_pat = "http://nusmv.fbk.eu/distrib/NuSMV-2.6.0-%s.tar.gz"
    url_pat = "https://loicpauleve.name/share/NuSMV-2.6.0-%s.tar.gz"
    binfile = {
        "linux64": "https://api.anaconda.org/download/colomoto/nusmv/2.6.0/linux-64/nusmv-2.6.0-0.tar.bz2",
        "linux32": url_pat % "linux32",
        "darwin": "https://api.anaconda.org/download/colomoto/nusmv/2.6.0/osx-64/nusmv-2.6.0-0.tar.bz2",
        "windows": url_pat % "win32",
    }
    binfile = binfile.get(system)
    def match_entry(m):
        return m.name.endswith("bin/NuSMV") \
            or  m.name.endswith("bin/NuSMV.exe")
    install_from_tarurl(binfile, match_entry, prefix)

setup({"pkg": "conda-forge/graphviz", "version": "2.38.0", "check_progs": ["dot"]},
        {"pkg": "colomoto/nusmv", "check_progs": ["NuSMV"], "install": nusmv_install})




# -*- coding: utf-8

from IPython.display import display, FileLink

import sys
import re

from .utils import *

modes = [
    "sync",
    "async",
    "general"
]


def bn2smv(bnfile, mode, init=None, name=None):
    assert mode in modes, "Le dernier argument doit être parmis %s" % " ".join(modes)

    hide = False
    if isinstance(bnfile, dict):
        defs = bnfile
        if not name:
            name = "tmpbn"
            hide = True
    else:
        name = name or bnfile.replace(".bn", "")
        defs = read_bn(bnfile)

    smvfile = "%s-%s.smv" % (name, mode)

    dom = list(sorted(defs.keys()))
    udom = ["u%s" % n for n in dom]

    def var(n):
        if isinstance(n, int):
            return "x%d" % n
        return n

    lines = ["MODULE main"]
    lines.append("VAR")
    for i in dom:
        lines.append("%s: boolean;" % var(i))
    lines.append("ASSIGN")
    for i in dom:
        if mode == "sync":
            lines.append("next(%s) := f%s;" % (var(i),i))
        else:
            lines.append("next(%s) := {%s, f%s};" % (var(i),var(i),i))
    lines.append("DEFINE")
    for d in sorted(defs.items()):
        lines.append("f%s := %s;" % d)
    if mode != "sync":
        lines.append("FIXEDPOINTS := %s;" \
            % (" & ".join(["%s = f%s" % (var(i),i) for i in dom])))
        lines.append("TRANS")
        lines.append("  FIXEDPOINTS")
        if mode == "general":
            for i in dom:
                lines.append("| next(%s) != %s" % (var(i),var(i)))
        elif mode == "async":
            for i in dom:
                freeze = " & ".join(["next({0})={0}".format(var(j)) for j in dom if i != j])
                freeze = " & %s" % freeze if freeze else ""
                lines.append("| next({0})!={0}{1}".format(var(i),freeze))
        lines.append(";")
    lines.append("")
    lines.append("-- INIT XXXX;")
    lines.append("-- CTLSPEC XXXX;")
    lines.append("-- LTLSPEC XXXX;")
    lines.append("")

    open(smvfile, "w").write("\n".join(lines))
    if not hide:
        display(FileLink(smvfile, url_prefix="../edit/"))

    if init is not None:
        for n in init:
            assert n in defs, "Nœud %s inconnu" % repr(n)
        smv_init = []
        for n in sorted(defs):
            smv_init.append("%s%s" % ("!" if n not in init else "" ,var(n)))
        append_line(smvfile, "INIT %s;" % (" & ".join(smv_init)))

    return smvfile






import itertools
import re

import pydotplus

from .utils import *

class state:
    def __init__(self, *t):
        self._d = t

    @classmethod
    def zero(celf, n):
        v = (False,)*n
        return celf(*v)

    def update(self, i, v):
        v = self[:i] + (v,) + self[i+1:]
        return state(*v)

    def __len__(self):
        return self._d.__len__()
    def __getitem__(self, i):
        return self._d[i]

    def next(self):
        x = self
        for i in range(len(self)):
            if not x[i]:
                return x.update(i, True)
            else:
                x = x.update(i, False)
        return x

    def iszero(self):
        return not (True in self._d)
    def isone(self):
        return not (False in self._d)

    def __hash__(self):
        return hash(self._d)
    def __eq__(self, x):
        return self._d == x._d
    def __ne__(self, x):
        return self._d != x._d
    def __le__(self, x):
        return self._d <= x._d
    def __lt__(self, x):
        return self._d < x._d
    def __ge__(self, x):
        return self._d >= x._d
    def __gt__(self, x):
        return self._d > x._d
    def __repr__(self):
        return str(self)
    def __str__(self):
        return "".join(["1" if i else "0" for i in self._d])
    def __iter__(self):
        return self._d.__iter__()

    @classmethod
    def all(celf, n):
        x = celf.zero(n)
        yield x
        while not x.isone():
            x = x.next()
            yield x

    def dst(x, y):
        d = 0
        for i in range(len(x)):
            if x[i] != y[i]:
                d += 1
        return d



class graph(dict):
    def __init__(self):
        super(graph, self).__init__()
        self.parents = {}

    def register_node(self, n):
        if n not in self:
            self[n] = set()
        if n not in self.parents:
            self.parents[n] = set()

    def add_edge(self, n1, n2):
        assert n1 != n2
        self.register_node(n1)
        self.register_node(n2)
        self[n1].add(n2)
        self.parents[n2].add(n1)

    def to_dot(self, dotfile):
        with open(dotfile, "w") as f:
            f.write("digraph {\n")
            for n, children in self.items():
                f.write("%s;\n"% n)
                for m in children:
                    f.write("%s -> %s;\n" % (n, m))
            f.write("}\n")

    def to_pydot(self):
        g = pydotplus.Dot('', graph_type="digraph", strict=True)
        for n, children in self.items():
            g.add_node(pydotplus.Node(str(n)))
            for m in children:
                edge = pydotplus.Edge(str(n), str(m))
                g.add_edge(edge)
        return g

    def tarjan(g):
        global num
        num = 0
        P = []
        ret = []

        nums = {}
        root = {}

        def parcours(v):
            global num
            nums[v] = num
            root[v] = num
            num += 1
            P.append(v)

            for w in g[v]:
                if w not in nums:
                    parcours(w)
                    root[v] = min(root[v], root[w])
                elif w in P:
                    root[v] = min(root[v], nums[w])
            if root[v] == nums[v]:
                scc = set()
                w = None
                while w is None or w != v:
                    w = P.pop()
                    scc.add(w)
                ret.append(scc)

        for v in g:
            if v not in nums:
                parcours(v)

        return ret


re_var = re.compile("\\b([\\w\\d_]+)\\b")
re_varidx = re.compile("^x(\\d+)$")

def bn2py(bn):
    if isinstance(bn, str):
        bn = read_bn(bn)

    nodes = list(sorted(bn))

    def get_index(m):
        i = m.group(1)
        if i in ["True", "False"]:
            return i
        m = re_varidx.match(i)
        if m:
            i = int(m.group(1))
        return "x[%s]" % nodes.index(i)

    exprs = []
    for n in nodes:
        d = bn[n]
        d = d.replace("TRUE", "True")
        d = d.replace("FALSE", "False")
        d = re_var.sub(get_index, d)
        d = d.replace("|", " or ")
        d = d.replace("&", " and ")
        d = d.replace("!", " not ")
        exprs.append("lambda x: %s" % d.strip())

    return eval("(%s)" % ",".join(exprs))


synchronous = "sync"
asynchronous = "async"
general = "general"

def stategraph(f, mode):
    if isinstance(f, (str,dict)):
        f = bn2py(f)
    n = len(f)
    sg = graph()
    dims = list(range(n))
    def transitions(x, m, M):
        for k in range(m,M+1):
            for I in itertools.combinations(dims, k):
                y = x
                for i in I:
                    y = y.update(i, f[i](x))
                if y != x:
                    sg.add_edge(x,y)
    for x in state.all(n):
        sg.register_node(x)
        if mode == asynchronous:
            transitions(x, 1, 1)
        elif mode == synchronous:
            transitions(x, n, n)
        elif mode == general:
            transitions(x, 1, n)
    return sg

def basins(g, a):
    b = {}
    for i, scc in enumerate(a):
        # i est le numero de l'attracteur scc
        # x0 est un etat de depart appartenant a l'attracteur
        x0 = tuple(scc)[0]
        b[x0] = {i}
        todo = [x0]
        visited = set(todo)
        while todo:
            x = todo.pop()
            for y in g.parents[x]:
                if y not in visited:
                    visited.add(y)
                    todo.append(y)
                    if y not in b:
                        b[y] = {i}
                    else:
                        b[y].add(i)
    return b

def get_attractor_id(b, x):
    assert len(b[x]) == 1, "%s can reach different attractors" % x
    return list(b[x])[0]





ERBB_BN = """fEGF := TRUE;
fERBB1 := EGF;
fERBB2 := EGF;
fERBB3 := EGF;
fERBB1_2 := ERBB1 & ERBB2;
fERBB1_3 := ERBB1 & ERBB3;
fERBB2_3 := ERBB2 & ERBB3;
fIGF1R := (ERa | AKT1) & !ERBB2_3;
fERa := AKT1 | MEK1;
fcMYC := AKT1 | MEK1 | ERa;
fAKT1 := ERBB1 | ERBB1_2 | ERBB1_3 | ERBB2_3 | IGF1R;
fMEK1 := ERBB1 | ERBB1_2 | ERBB1_3 | ERBB2_3 | IGF1R;
fCDK2 := CycE1 & !p21 & !p27;
fCDK4 := CycD1 & !p21 & !p27;
fCDK6 := CycD1;
fCycD1 := AKT1 | MEK1 | ERa | cMYC;
fCycE1 := cMYC;
fp21 := ERa & !AKT1 & !cMYC & !CDK4;
fp27 := ERa & !CDK4 & !CDK2 & !AKT1 & !cMYC;
fpRB := (CDK4 & CDK6);"""




# -*- coding: utf-8

from IPython.display import display, FileLink

import re

def create_file(filename, content):
    with open(filename, "w") as f:
        f.write(content)
    display(FileLink(filename, url_prefix="../edit/"))
    return filename

def append_line(filename, line, quiet=True):
    with open(filename, "a") as f:
        f.write("%s\n" % line)
    if not quiet:
        print("Ligne '%s' ajoutée à la fin du fichier %s" % (line, filename))

re_fline = re.compile("\\bf([\\w\\d_]+)\\s*:=\\s*([\\sx\\w\\d_+&!|()TRUEFALSE]+)\s*;")
re_int = re.compile("^\\d+$")

def parse_bn(content):
    content = re.sub("\\s","",content.strip())
    defs = re_fline.findall(content)
    data = "".join(["f%s:=%s;" % d for d in defs])
    if data != content:
        raise Exception("Invalid syntax")

    def typekey(kv):
        (k,v) = kv
        if re_int.match(k):
            k = int(k)
        return (k,v)

    return dict(map(typekey, defs))

def read_bn(bnfile):
    with open(bnfile) as f:
        return parse_bn(f.read())





from setuptools import setup, find_packages

NAME = "modian"

setup(name=NAME,
    version='0.3.5',
    author = "Loïc Paulevé",
    author_email = "loic.pauleve@labri.fr",
    description = "Simple python library for Boolean networks",
    long_description = "For teaching purpose",
    install_requires = [
        "jupyter",
        "notebook",
        "pydotplus"
    ],
    py_modules = ["modian_setup"],
    license="CeCILL",
    packages = [NAME],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="boolean networks, model checing, teaching",
)





from IPython.display import *

import os
import platform
import subprocess


from .utils import *
from .bnutils import *
from .bn2smv import *
from .models import *

##
# Setup PATH env for custom modian installations
#
if platform.system() == "Windows":
    PREFIXES = [os.path.join(os.getenv("APPDATA"), "modian")]
    binpaths = [os.path.join(p,"bin") for p in PREFIXES] \
                + [os.path.join(p,"Library","bin") for p in PREFIXES] \
                + [os.path.join(p,"Library","bin", "graphviz") for p in PREFIXES]
else:
    PREFIXES = [os.path.join(os.path.expanduser("~"), ".local", "share", "modian")]
    binpaths = [os.path.join(p, "bin") for p in PREFIXES]

binpaths = [p for p in binpaths if os.path.exists(p)]
if binpaths:
    os.environ["PATH"] = "%s%s%s" % (os.pathsep.join(binpaths), os.pathsep, os.environ["PATH"])
#
##


PYTHON3 = sys.version_info[0] >= 3

class CmdError(subprocess.CalledProcessError):
    def __str__(self):
        stderr = "\n%s" % self.stderr.decode() if hasattr(self, "stderr") and self.stderr \
            else self.output
        return "Command '%s' returned non-zero exit status %d\n%s\n" \
            % (" ".join(self.cmd), self.returncode, stderr)

def NuSMV(*args, **kwargs):
    quiet = kwargs.get("quiet", True)
    parse = kwargs.get("parse", False)
    parse_dict = kwargs.get("parse_dict", False)

    args = ["NuSMV"] + ["-dcx"] + list(args)
    try:
        stderr = subprocess.PIPE if PYTHON3 else subprocess.STDOUT
        output = subprocess.check_output(args, stderr=stderr).decode().strip()
    except subprocess.CalledProcessError as e:
        # backward compatible 'raise e from None'
        if PYTHON3:
            e = CmdError(e.returncode, e.cmd, e.output, e.stderr)
        else:
            e = CmdError(e.returncode, e.cmd, e.output)
        e.__cause__ = None
        raise e
    ret = None
    if quiet or parse:
        lines = output.split("\n")
        if quiet:
            lines = [l for l in lines if l \
                    and not l.startswith("WARNING") \
                    and not l.startswith("***")]
        if parse:
            if parse_dict:
                ret = {}
            else:
                ret = []
            for line in lines:
                if not line.startswith("-- specification "):
                    continue
                parts = line.split()
                result = True if parts[-1] == "true" else False
                if parse_dict:
                    prop = " ".join(parts[2:-2])
                    ret[prop] = result
                else:
                    ret.append(result)
        output = "\n".join(lines)
    if not parse or not quiet:
        print(output)
    return ret


def svg_of_graph(g):
    return g.to_pydot().create_svg().decode()

def setup_ipython():
    import pydotplus
    try:
        ip = get_ipython()
    except NameError:
        return
    # nxgraph to svg
    svg_formatter = ip.display_formatter.formatters["image/svg+xml"]
    svg_formatter.for_type(graph, svg_of_graph)
    svg_formatter.for_type(pydotplus.Dot, lambda g: g.create_svg().decode())

setup_ipython()

